import dataclasses
import json
from typing import Any, Dict, List, Optional, TypedDict

from qstash.http import HttpClient
from qstash.message import Message, parse_flow_control


@dataclasses.dataclass
class DlqMessage(Message):
    dlq_id: str
    """Unique id within the DLQ."""

    response_status: int
    """HTTP status code of the last failed delivery attempt."""

    response_headers: Optional[Dict[str, List[str]]]
    """Response headers of the last failed delivery attempt."""

    response_body: Optional[str]
    """
    Response body of the last failed delivery attempt if it is
    composed of UTF-8 characters only.
    """

    response_body_base64: Optional[str]
    """
    Base64 encoded response body of the last failed delivery attempt
    if the response body contains non-UTF-8 characters.
    """

    retry_delay_expression: Optional[str]
    """
    The retry delay expression for this DLQ message,
    if retry_delay was set when publishing the message.
    """


class DlqFilter(TypedDict, total=False):
    message_id: str
    """Filter DLQ entries by message id."""

    url: str
    """Filter DLQ entries by url."""

    url_group: str
    """Filter DLQ entries by url group name."""

    queue: str
    """Filter DLQ entries by queue name."""

    schedule_id: str
    """Filter DLQ entries by schedule id."""

    from_time: int
    """Filter DLQ entries by starting Unix time, in milliseconds"""

    to_time: int
    """Filter DLQ entries by ending Unix time, in milliseconds"""

    response_status: int
    """Filter DLQ entries by HTTP status of the response"""

    caller_ip: str
    """Filter DLQ entries by IP address of the publisher of the message"""

    label: str
    """Filter DLQ entries by label."""


@dataclasses.dataclass
class ListDlqMessagesResponse:
    cursor: Optional[str]
    """
    A cursor which can be used in subsequent requests to paginate through
    all messages. If `None`, end of the DLQ messages are reached.
    """

    messages: List[DlqMessage]
    """List of DLQ messages."""


def parse_dlq_message_response(
    response: Dict[str, Any],
    dlq_id: str = "",
) -> DlqMessage:
    flow_control = parse_flow_control(response)
    return DlqMessage(
        message_id=response["messageId"],
        url=response["url"],
        url_group=response.get("topicName"),
        endpoint=response.get("endpointName"),
        queue=response.get("queueName"),
        body=response.get("body"),
        body_base64=response.get("bodyBase64"),
        method=response["method"],
        headers=response.get("header"),
        callback_headers=response.get("callbackHeader"),
        failure_callback_headers=response.get("failureCallbackHeader"),
        max_retries=response["maxRetries"],
        not_before=response["notBefore"],
        created_at=response["createdAt"],
        callback=response.get("callback"),
        failure_callback=response.get("failureCallback"),
        schedule_id=response.get("scheduleId"),
        caller_ip=response.get("callerIP"),
        dlq_id=response.get("dlqId", dlq_id),
        response_status=response["responseStatus"],
        response_headers=response.get("responseHeader"),
        response_body=response.get("responseBody"),
        response_body_base64=response.get("responseBodyBase64"),
        flow_control=flow_control,
        retry_delay_expression=response.get("retryDelayExpression"),
        label=response.get("label"),
    )


def prepare_list_dlq_messages_params(
    *,
    cursor: Optional[str],
    count: Optional[int],
    filter: Optional[DlqFilter],
) -> Dict[str, str]:
    params = {}

    if cursor is not None:
        params["cursor"] = cursor

    if count is not None:
        params["count"] = str(count)

    if filter is not None:
        if "message_id" in filter:
            params["messageId"] = filter["message_id"]

        if "url" in filter:
            params["url"] = filter["url"]

        if "url_group" in filter:
            params["topicName"] = filter["url_group"]

        if "queue" in filter:
            params["queueName"] = filter["queue"]

        if "schedule_id" in filter:
            params["scheduleId"] = filter["schedule_id"]

        if "from_time" in filter:
            params["fromDate"] = str(filter["from_time"])

        if "to_time" in filter:
            params["toDate"] = str(filter["to_time"])

        if "response_status" in filter:
            params["responseStatus"] = str(filter["response_status"])

        if "caller_ip" in filter:
            params["callerIp"] = filter["caller_ip"]

        if "label" in filter:
            params["label"] = filter["label"]

    return params


class DlqApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def get(self, dlq_id: str) -> DlqMessage:
        """
        Gets a message from DLQ.

        :param dlq_id: The unique id within the DLQ to get.
        """
        response = self._http.request(
            path=f"/v2/dlq/{dlq_id}",
            method="GET",
        )

        return parse_dlq_message_response(response, dlq_id)

    def list(
        self,
        *,
        cursor: Optional[str] = None,
        count: Optional[int] = None,
        filter: Optional[DlqFilter] = None,
    ) -> ListDlqMessagesResponse:
        """
        Lists all messages currently inside the DLQ.

        :param cursor: Optional cursor to start listing DLQ messages from.
        :param count: The maximum number of DLQ messages to return.
            Default and max is `100`.
        :param filter: Filter to use.
        """
        params = prepare_list_dlq_messages_params(
            cursor=cursor,
            count=count,
            filter=filter,
        )

        response = self._http.request(
            path="/v2/dlq",
            method="GET",
            params=params,
        )

        messages = [parse_dlq_message_response(r) for r in response["messages"]]

        return ListDlqMessagesResponse(
            cursor=response.get("cursor"),
            messages=messages,
        )

    def delete(self, dlq_id: str) -> None:
        """
        Deletes a message from the DLQ.

        :param dlq_id: The unique id within the DLQ to delete.
        """
        self._http.request(
            path=f"/v2/dlq/{dlq_id}",
            method="DELETE",
            parse_response=False,
        )

    def delete_many(self, dlq_ids: List[str]) -> int:
        """
        Deletes multiple messages from the DLQ and
        returns how many of them are deleted.

        :param dlq_ids: The unique ids within the DLQ to delete.
        """
        body = json.dumps({"dlqIds": dlq_ids})

        response = self._http.request(
            path="/v2/dlq",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
        )

        return response["deleted"]  # type:ignore[no-any-return]



import json
from typing import Any, Dict, List, Optional, Union

from qstash.asyncio.http import AsyncHttpClient
from qstash.http import HttpMethod
from qstash.message import (
    ApiT,
    FlowControl,
    Redact,
    BatchJsonRequest,
    BatchRequest,
    BatchResponse,
    BatchUrlGroupResponse,
    EnqueueResponse,
    EnqueueUrlGroupResponse,
    Message,
    PublishResponse,
    PublishUrlGroupResponse,
    convert_to_batch_messages,
    get_destination,
    parse_batch_response,
    parse_enqueue_response,
    parse_message_response,
    parse_publish_response,
    prepare_batch_message_body,
    prepare_headers,
)


class AsyncMessageApi:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def publish(
        self,
        *,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        not_before: Optional[int] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[PublishResponse, List[PublishUrlGroupResponse]]:
        """
        Publishes a message to QStash.

        If the destination is a `url` or an `api`, `PublishResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `PublishUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param not_before: Delay the message until a certain time in the future.
            The format is a unix timestamp in seconds, based on the UTC timezone.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        """
        headers = headers or {}
        destination = get_destination(
            url=url,
            url_group=url_group,
            api=api,
            headers=headers,
        )

        req_headers = prepare_headers(
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            not_before=not_before,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

        response = await self._http.request(
            path=f"/v2/publish/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return parse_publish_response(response)

    async def publish_json(
        self,
        *,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        not_before: Optional[int] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[PublishResponse, List[PublishUrlGroupResponse]]:
        """
        Publish a message to QStash, automatically serializing the
        body as JSON string, and setting content type to `application/json`.

        If the destination is a `url` or an `api`, `PublishResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `PublishUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param not_before: Delay the message until a certain time in the future.
            The format is a unix timestamp in seconds, based on the UTC timezone.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        """
        return await self.publish(
            url=url,
            url_group=url_group,
            api=api,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            not_before=not_before,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

    async def enqueue(
        self,
        *,
        queue: str,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[EnqueueResponse, List[EnqueueUrlGroupResponse]]:
        """
        Enqueues a message, after creating the queue if it does
        not exist.

        If the destination is a `url` or an `api`, `EnqueueResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `EnqueueUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param queue: The name of the queue.
        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param label: Assign a label to the request to filter logs with it later.
        """
        headers = headers or {}
        destination = get_destination(
            url=url,
            url_group=url_group,
            api=api,
            headers=headers,
        )

        req_headers = prepare_headers(
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=None,
            not_before=None,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=None,
            label=label,
            redact=redact,
        )

        response = await self._http.request(
            path=f"/v2/enqueue/{queue}/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return parse_enqueue_response(response)

    async def enqueue_json(
        self,
        *,
        queue: str,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[EnqueueResponse, List[EnqueueUrlGroupResponse]]:
        """
        Enqueues a message, after creating the queue if it does
        not exist. It automatically serializes the body as JSON string,
        and setting content type to `application/json`.

        If the destination is a `url` or an `api`, `EnqueueResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `EnqueueUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param queue: The name of the queue.
        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param label: Assign a label to the request to filter logs with it later.
        """
        return await self.enqueue(
            queue=queue,
            url=url,
            url_group=url_group,
            api=api,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            label=label,
            redact=redact,
        )

    async def batch(
        self, messages: List[BatchRequest]
    ) -> List[Union[BatchResponse, List[BatchUrlGroupResponse]]]:
        """
        Publishes or enqueues multiple messages in a single request.

        Returns a list of publish or enqueue responses, one for each
        message in the batch.

        If the message in the batch is sent to a url or an API,
        the corresponding item in the response is `BatchResponse`.

        If the message in the batch is sent to a url group,
        the corresponding item in the response is list of
        `BatchUrlGroupResponse`s, one for each url in the url group.
        """
        body = prepare_batch_message_body(messages)

        response = await self._http.request(
            path="/v2/batch",
            body=body,
            headers={"Content-Type": "application/json"},
            method="POST",
        )

        return parse_batch_response(response)

    async def batch_json(
        self, messages: List[BatchJsonRequest]
    ) -> List[Union[BatchResponse, List[BatchUrlGroupResponse]]]:
        """
        Publishes or enqueues multiple messages in a single request,
        automatically serializing the message bodies as JSON strings,
        and setting content type to `application/json`.

        Returns a list of publish or enqueue responses, one for each
        message in the batch.

        If the message in the batch is sent to a url or an API,
        the corresponding item in the response is `BatchResponse`.

        If the message in the batch is sent to a url group,
        the corresponding item in the response is list of
        `BatchUrlGroupResponse`s, one for each url in the url group.
        """
        batch_messages = convert_to_batch_messages(messages)
        return await self.batch(batch_messages)

    async def get(self, message_id: str) -> Message:
        """
        Gets the message by its id.
        """
        response = await self._http.request(
            path=f"/v2/messages/{message_id}",
            method="GET",
        )

        return parse_message_response(response)

    async def cancel(self, message_id: str) -> None:
        """
        Cancels delivery of an existing message.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.
        """
        await self._http.request(
            path=f"/v2/messages/{message_id}",
            method="DELETE",
            parse_response=False,
        )

    async def cancel_many(self, message_ids: List[str]) -> int:
        """
        Cancels delivery of existing messages.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.

        Returns how many of the messages are cancelled.
        """
        body = json.dumps({"messageIds": message_ids})

        response = await self._http.request(
            path="/v2/messages",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
        )

        return response["cancelled"]  # type:ignore[no-any-return]

    async def cancel_all(self) -> int:
        """
        Cancels delivery of all the existing messages.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.

        Returns how many messages are cancelled.
        """
        response = await self._http.request(
            path="/v2/messages",
            method="DELETE",
        )

        return response["cancelled"]  # type:ignore[no-any-return]



import dataclasses
import enum
from typing import Any, Dict, List, Optional, TypedDict

from qstash.http import HttpClient, HttpMethod
from qstash.message import parse_flow_control, FlowControlProperties


class LogState(enum.Enum):
    """State of the message."""

    CREATED = "CREATED"
    """Message has been accepted and stored in QStash"""

    ACTIVE = "ACTIVE"
    """Task is currently being processed by a worker."""

    RETRY = "RETRY"
    """Task has been scheduled to retry."""

    ERROR = "ERROR"
    """
    Execution threw an error and the task is waiting to be retried
    or failed.
    """

    DELIVERED = "DELIVERED"
    """Message was successfully delivered."""

    FAILED = "FAILED"
    """
    Task has failed too many times or encountered an error that it
    cannot recover from.
    """

    CANCEL_REQUESTED = "CANCEL_REQUESTED"
    """Cancel request from the user is recorded."""

    CANCELED = "CANCELED"
    """Cancel request from the user is honored."""

    IN_PROGRESS = "IN_PROGRESS"
    """Messages which are in progress"""


@dataclasses.dataclass
class Log:
    time: int
    """Unix time of the log entry, in milliseconds."""

    message_id: str
    """Message id associated with the log."""

    state: LogState
    """Current state of the message at this point in time."""

    error: Optional[str]
    """An explanation what went wrong."""

    next_delivery_time: Optional[int]
    """Next scheduled Unix time of the message, milliseconds."""

    url: str
    """Destination url."""

    url_group: Optional[str]
    """Name of the url group if this message was sent through a url group."""

    endpoint: Optional[str]
    """Name of the endpoint if this message was sent through a url group."""

    queue: Optional[str]
    """Name of the queue if this message is enqueued on a queue."""

    schedule_id: Optional[str]
    """Schedule id of the message if the message is triggered by a schedule."""

    body_base64: Optional[str]
    """Base64 encoded body of the message."""

    headers: Optional[Dict[str, List[str]]]
    """Headers of the message"""

    callback_headers: Optional[Dict[str, List[str]]]
    """Headers of the callback message"""

    failure_callback_headers: Optional[Dict[str, List[str]]]
    """Headers of the failure callback message"""

    response_status: Optional[int]
    """HTTP status code of the last failed delivery attempt."""

    response_headers: Optional[Dict[str, List[str]]]
    """Response headers of the last failed delivery attempt."""

    response_body: Optional[str]
    """Response body of the last failed delivery attempt."""

    timeout: Optional[int]
    """HTTP timeout value used while calling the destination url."""

    method: Optional[HttpMethod]
    """HTTP method to use to deliver the message."""

    callback: Optional[str]
    """Url which is called each time the message is attempted to be delivered."""

    failure_callback: Optional[str]
    """Url which is called after the message is failed."""

    max_retries: Optional[int]
    """Number of retries that should be attempted in case of delivery failure."""

    retry_delay_expression: Optional[str]
    """
    The retry delay expression for this DLQ message,
    if retry_delay was set when publishing the message.
    """

    flow_control: Optional[FlowControlProperties]
    """Flow control properties"""

    label: Optional[str]
    """Label assigned to the request for filtering logs."""


class LogFilter(TypedDict, total=False):
    message_id: str
    """Filter logs by message id."""

    message_ids: List[str]
    """Filter logs by message ids."""

    state: LogState
    """Filter logs by state."""

    url: str
    """Filter logs by url."""

    url_group: str
    """Filter logs by url group name."""

    queue: str
    """Filter logs by queue name."""

    schedule_id: str
    """Filter logs by schedule id."""

    from_time: int
    """Filter logs by starting Unix time, in milliseconds"""

    to_time: int
    """Filter logs by ending Unix time, in milliseconds"""

    label: str
    """Filter logs by label."""


@dataclasses.dataclass
class ListLogsResponse:
    cursor: Optional[str]
    """
    A cursor which can be used in subsequent requests to paginate through 
    all logs. If `None`, end of the logs are reached.
    """

    logs: List[Log]
    """List of logs."""


def prepare_list_logs_request_params(
    *,
    cursor: Optional[str],
    count: Optional[int],
    filter: Optional[LogFilter],
) -> Dict[str, str]:
    params: Dict[str, Any] = {}

    if cursor is not None:
        params["cursor"] = cursor

    if count is not None:
        params["count"] = str(count)

    if filter is not None:
        if "message_id" in filter:
            params["messageId"] = filter["message_id"]

        if "message_ids" in filter:
            params["messageIds"] = filter["message_ids"]

        if "state" in filter:
            params["state"] = filter["state"].value

        if "url" in filter:
            params["url"] = filter["url"]

        if "url_group" in filter:
            params["topicName"] = filter["url_group"]

        if "queue" in filter:
            params["queueName"] = filter["queue"]

        if "schedule_id" in filter:
            params["scheduleId"] = filter["schedule_id"]

        if "from_time" in filter:
            params["fromDate"] = str(filter["from_time"])

        if "to_time" in filter:
            params["toDate"] = str(filter["to_time"])

        if "label" in filter:
            params["label"] = filter["label"]

    return params


def parse_logs_response(response: List[Dict[str, Any]]) -> List[Log]:
    logs = []

    for event in response:
        flow_control = parse_flow_control(event)
        logs.append(
            Log(
                time=event["time"],
                message_id=event["messageId"],
                state=LogState(event["state"]),
                error=event.get("error"),
                next_delivery_time=event.get("nextDeliveryTime"),
                url=event["url"],
                url_group=event.get("topicName"),
                endpoint=event.get("endpointName"),
                queue=event.get("queueName"),
                schedule_id=event.get("scheduleId"),
                headers=event.get("header"),
                callback_headers=event.get("callbackHeaders"),
                failure_callback_headers=event.get("failureCallbackHeaders"),
                body_base64=event.get("body"),
                response_status=event.get("responseStatus"),
                response_headers=event.get("responseHeader"),
                response_body=event.get("responseBody"),
                timeout=event.get("timeout"),
                callback=event.get("callback"),
                failure_callback=event.get("failureCallback"),
                flow_control=flow_control,
                method=event.get("method"),
                max_retries=event.get("maxRetries"),
                retry_delay_expression=event.get("retryDelayExpr"),
                label=event.get("label"),
            )
        )

    return logs


class LogApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def list(
        self,
        *,
        cursor: Optional[str] = None,
        count: Optional[int] = None,
        filter: Optional[LogFilter] = None,
    ) -> ListLogsResponse:
        """
        Lists all logs that happened, such as message creation or delivery.

        :param cursor: Optional cursor to start listing logs from.
        :param count: The maximum number of logs to return.
            Default and max is `1000`.
        :param filter: Filter to use.
        """
        params = prepare_list_logs_request_params(
            cursor=cursor,
            count=count,
            filter=filter,
        )

        response = self._http.request(
            path="/v2/events",
            method="GET",
            params=params,
        )

        print(response)

        logs = parse_logs_response(response["events"])

        return ListLogsResponse(
            cursor=response.get("cursor"),
            logs=logs,
        )



import math
import time
from typing import TypedDict, Callable, Optional, Union, Literal, Any, Dict

import httpx

from qstash.errors import (
    RateLimitExceededError,
    QStashError,
    DailyMessageLimitExceededError,
)


class RetryConfig(TypedDict, total=False):
    retries: int
    """Maximum number of retries will be performed after the initial request fails."""

    backoff: Callable[[int], float]
    """A function that returns how many milliseconds to backoff before the given retry attempt."""


DEFAULT_TIMEOUT = httpx.Timeout(
    timeout=600.0,
    connect=5.0,
)

DEFAULT_RETRY = RetryConfig(
    retries=5,
    backoff=lambda retry_count: math.exp(1 + retry_count) * 50,
)

NO_RETRY = RetryConfig(
    retries=0,
    backoff=lambda _: 0,
)

BASE_URL = "https://qstash.upstash.io"

HttpMethod = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]


def daily_message_limit_error(headers: httpx.Headers) -> DailyMessageLimitExceededError:
    limit = headers.get("RateLimit-Limit")
    remaining = headers.get("RateLimit-Remaining")
    reset = headers.get("RateLimit-Reset")
    return DailyMessageLimitExceededError(
        limit=limit,
        remaining=remaining,
        reset=reset,
    )


def burst_rate_limit_error(headers: httpx.Headers) -> RateLimitExceededError:
    limit = headers.get("Burst-RateLimit-Limit")
    remaining = headers.get("Burst-RateLimit-Remaining")
    reset = headers.get("Burst-RateLimit-Reset")
    return RateLimitExceededError(
        limit=limit,
        remaining=remaining,
        reset=reset,
    )


def raise_for_non_ok_status(response: httpx.Response) -> None:
    if response.is_success:
        return

    if response.status_code == 429:
        headers = response.headers
        if "RateLimit-Limit" in headers:
            raise daily_message_limit_error(headers)
        else:
            raise burst_rate_limit_error(headers)

    raise QStashError(
        f"Request failed with status: {response.status_code}, body: {response.text}"
    )


class HttpClient:
    def __init__(
        self,
        token: str,
        retry: Optional[Union[Literal[False], RetryConfig]],
        base_url: Optional[str] = None,
    ) -> None:
        self._token = f"Bearer {token}"

        if retry is None:
            self._retry = DEFAULT_RETRY
        elif retry is False:
            self._retry = NO_RETRY
        else:
            self._retry = retry

        self._client = httpx.Client(
            timeout=DEFAULT_TIMEOUT,
        )

        self._base_url = base_url.rstrip("/") if base_url else BASE_URL

    def request(
        self,
        *,
        path: str,
        method: HttpMethod,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Union[str, bytes]] = None,
        params: Optional[Dict[str, str]] = None,
        parse_response: bool = True,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Any:
        base_url = base_url or self._base_url
        token = token or self._token

        url = base_url + path
        headers = {"Authorization": token, **(headers or {})}

        max_attempts = 1 + max(0, self._retry["retries"])
        last_error = None
        response = None
        for attempt in range(max_attempts):
            try:
                response = self._client.request(
                    method=method,
                    url=url,
                    params=params,
                    headers=headers,
                    content=body,
                )
                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e
                backoff = self._retry["backoff"](attempt) / 1000
                time.sleep(backoff)

        if not response:
            # Can't be None at this point
            raise last_error  # type:ignore[misc]

        raise_for_non_ok_status(response)

        if parse_response:
            return response.json()

        return response.text

    def stream(
        self,
        *,
        path: str,
        method: HttpMethod,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Union[str, bytes]] = None,
        params: Optional[Dict[str, str]] = None,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
    ) -> httpx.Response:
        base_url = base_url or self._base_url
        token = token or self._token

        url = base_url + path
        headers = {"Authorization": token, **(headers or {})}

        max_attempts = 1 + max(0, self._retry["retries"])
        last_error = None
        response = None
        for attempt in range(max_attempts):
            try:
                request = self._client.build_request(
                    method=method,
                    url=url,
                    params=params,
                    headers=headers,
                    content=body,
                )
                response = self._client.send(
                    request,
                    stream=True,
                )
                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e
                backoff = self._retry["backoff"](attempt) / 1000
                time.sleep(backoff)

        if not response:
            # Can't be None at this point
            raise last_error  # type:ignore[misc]

        try:
            raise_for_non_ok_status(response)
        except Exception as e:
            response.close()
            raise e

        return response



import dataclasses
import json
from typing import Any, Dict, List

from qstash.http import HttpClient


@dataclasses.dataclass
class Queue:
    name: str
    """Name of the queue."""

    parallelism: int
    """Number of parallel consumers consuming from the queue."""

    created_at: int
    """Creation time of the queue, in Unix milliseconds."""

    updated_at: int
    """Last update time of the queue, in Unix milliseconds."""

    lag: int
    """Number of unprocessed messages that exist in the queue."""

    paused: bool
    """Whether the queue is paused or not."""


def prepare_upsert_body(queue: str, parallelism: int, paused: bool) -> str:
    return json.dumps(
        {
            "queueName": queue,
            "parallelism": parallelism,
            "paused": paused,
        }
    )


def parse_queue_response(response: Dict[str, Any]) -> Queue:
    return Queue(
        name=response["name"],
        parallelism=response["parallelism"],
        created_at=response["createdAt"],
        updated_at=response["updatedAt"],
        lag=response["lag"],
        paused=response["paused"],
    )


class QueueApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def upsert(
        self,
        queue: str,
        *,
        parallelism: int = 1,
        paused: bool = False,
    ) -> None:
        """
        Updates or creates a queue.

        :param queue: The name of the queue.
        :param parallelism: The number of parallel consumers consuming from the queue.
        :param paused: Whether to pause the queue or not. A paused queue will not
            deliver new messages until it is resumed.
        """
        body = prepare_upsert_body(queue, parallelism, paused)

        self._http.request(
            path="/v2/queues",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    def get(self, queue: str) -> Queue:
        """
        Gets the queue by its name.
        """
        response = self._http.request(
            path=f"/v2/queues/{queue}",
            method="GET",
        )

        return parse_queue_response(response)

    def list(self) -> List[Queue]:
        """
        Lists all the queues.
        """
        response = self._http.request(
            path="/v2/queues",
            method="GET",
        )

        return [parse_queue_response(r) for r in response]

    def delete(self, queue: str) -> None:
        """
        Deletes the queue.
        """
        self._http.request(
            path=f"/v2/queues/{queue}",
            method="DELETE",
            parse_response=False,
        )

    def pause(self, queue: str) -> None:
        """
        Pauses the queue.

        A paused queue will not deliver messages until
        it is resumed.
        """
        self._http.request(
            path=f"/v2/queues/{queue}/pause",
            method="POST",
            parse_response=False,
        )

    def resume(self, queue: str) -> None:
        """
        Resumes the queue.
        """
        self._http.request(
            path=f"/v2/queues/{queue}/resume",
            method="POST",
            parse_response=False,
        )



import dataclasses
import re


@dataclasses.dataclass
class LlmProvider:
    name: str
    """Name of the LLM provider."""

    base_url: str
    """Base URL of the provider."""

    token: str
    """
    The token for the provider.
    
    The provided key will be passed to the
    endpoint as a bearer token.
    """


def openai(token: str) -> LlmProvider:
    return LlmProvider(
        name="OpenAI",
        base_url="https://api.openai.com",
        token=token,
    )


def custom(base_url: str, token: str) -> LlmProvider:
    base_url = re.sub("/(v1/)?chat/completions$", "", base_url)
    return LlmProvider(
        name="custom",
        base_url=base_url,
        token=token,
    )



import dataclasses
from typing import Any, Dict, TypedDict

from qstash.http import HttpClient


@dataclasses.dataclass
class FlowControlInfo:
    """Information about a flow control key."""

    key: str
    """The flow control key."""

    wait_list_size: int
    """The number of messages waiting in the wait list."""

    parallelism_max: int
    """The maximum parallelism configured for this flow control key."""

    parallelism_count: int
    """The current number of active requests for this flow control key."""

    rate_max: int
    """The maximum rate configured for this flow control key."""

    rate_count: int
    """The current number of requests consumed in the current period."""

    rate_period: int
    """The rate period in seconds."""

    rate_period_start: int
    """The start time of the current rate period as a unix timestamp."""

    is_paused: bool
    """Whether message delivery is paused for this flow control key."""

    is_pinned_parallelism: bool
    """Whether the parallelism configuration is pinned."""

    is_pinned_rate: bool
    """Whether the rate configuration is pinned."""


@dataclasses.dataclass
class GlobalParallelismInfo:
    """Information about global parallelism."""

    parallelism_max: int
    """The maximum global parallelism."""

    parallelism_count: int
    """The current number of active requests globally."""


class PinFlowControlOptions(TypedDict, total=False):
    """Options for pinning a flow control key configuration."""

    parallelism: int
    """The parallelism value to apply to the flow-control key."""

    rate: int
    """The rate value to apply to the flow-control key."""

    period: int
    """The period value to apply to the flow-control key, in seconds."""


class UnpinFlowControlOptions(TypedDict, total=False):
    """Options for unpinning a flow control key configuration."""

    parallelism: bool
    """Whether to unpin the parallelism configuration."""

    rate: bool
    """Whether to unpin the rate configuration."""


def parse_flow_control_info(response: Dict[str, Any]) -> FlowControlInfo:
    return FlowControlInfo(
        key=response["flowControlKey"],
        wait_list_size=response.get("waitListSize", 0),
        parallelism_max=response.get("parallelismMax", 0),
        parallelism_count=response.get("parallelismCount", 0),
        rate_max=response.get("rateMax", 0),
        rate_count=response.get("rateCount", 0),
        rate_period=response.get("ratePeriod", 0),
        rate_period_start=response.get("ratePeriodStart", 0),
        is_paused=response.get("isPaused", False),
        is_pinned_parallelism=response.get("isPinnedParallelism", False),
        is_pinned_rate=response.get("isPinnedRate", False),
    )


class FlowControlApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def get(self, flow_control_key: str) -> FlowControlInfo:
        """
        Gets a single flow control by key.

        :param flow_control_key: The flow control key to get.
        """
        response = self._http.request(
            path=f"/v2/flowControl/{flow_control_key}",
            method="GET",
        )

        return parse_flow_control_info(response)

    def get_global_parallelism(self) -> GlobalParallelismInfo:
        """
        Gets the global parallelism info.
        """
        response = self._http.request(
            path="/v2/globalParallelism",
            method="GET",
        )

        return GlobalParallelismInfo(
            parallelism_max=response.get("parallelismMax", 0),
            parallelism_count=response.get("parallelismCount", 0),
        )

    def pause(self, flow_control_key: str) -> None:
        """
        Pause message delivery for a flow-control key.

        Messages already in the waitlist will remain there.
        New incoming messages will be added directly to the waitlist.

        :param flow_control_key: The flow control key to pause.
        """
        self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/pause",
            method="POST",
            parse_response=False,
        )

    def resume(self, flow_control_key: str) -> None:
        """
        Resume message delivery for a flow-control key.

        :param flow_control_key: The flow control key to resume.
        """
        self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/resume",
            method="POST",
            parse_response=False,
        )

    def pin(self, flow_control_key: str, options: PinFlowControlOptions) -> None:
        """
        Pin a processing configuration for a flow-control key.

        While pinned, the system ignores configurations provided by incoming
        messages and uses the pinned configuration instead.

        :param flow_control_key: The flow control key to pin.
        :param options: The configuration to pin.
        """
        params: Dict[str, str] = {}
        if "parallelism" in options:
            params["parallelism"] = str(options["parallelism"])
        if "rate" in options:
            params["rate"] = str(options["rate"])
        if "period" in options:
            params["period"] = str(options["period"])

        self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/pin",
            method="POST",
            params=params or None,
            parse_response=False,
        )

    def unpin(self, flow_control_key: str, options: UnpinFlowControlOptions) -> None:
        """
        Remove the pinned configuration for a flow-control key.

        After unpinning, the system resumes updating the configuration
        based on incoming messages.

        :param flow_control_key: The flow control key to unpin.
        :param options: Which configurations to unpin.
        """
        params: Dict[str, str] = {}
        if "parallelism" in options:
            params["parallelism"] = str(options["parallelism"]).lower()
        if "rate" in options:
            params["rate"] = str(options["rate"]).lower()

        self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/unpin",
            method="POST",
            params=params or None,
            parse_response=False,
        )

    def reset_rate(self, flow_control_key: str) -> None:
        """
        Reset the rate configuration state for a flow-control key.

        Clears the current rate count and immediately ends the current period.
        The current timestamp becomes the start of the new rate period.

        :param flow_control_key: The flow control key to reset rate for.
        """
        self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/resetRate",
            method="POST",
            parse_response=False,
        )



import json
from typing import Any, Dict, List, Optional, Union

from qstash.asyncio.http import AsyncHttpClient
from qstash.http import HttpMethod
from qstash.message import FlowControl, Redact
from qstash.schedule import (
    Schedule,
    parse_schedule_response,
    prepare_schedule_headers,
)


class AsyncScheduleApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def create(
        self,
        *,
        destination: str,
        cron: str,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        timeout: Optional[Union[str, int]] = None,
        schedule_id: Optional[str] = None,
        queue: Optional[str] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> str:
        """
        Creates a schedule to send messages periodically.

        Returns the created schedule id.

        :param destination: The destination url or url group.
        :param cron: The cron expression to use to schedule the messages.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param schedule_id: Schedule id to use. This can be used to update the settings
            of an existing schedule.
        :param queue: Name of the queue which the scheduled messages will be enqueued.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        """
        req_headers = prepare_schedule_headers(
            cron=cron,
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            timeout=timeout,
            schedule_id=schedule_id,
            queue=queue,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

        response = await self._http.request(
            path=f"/v2/schedules/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return response["scheduleId"]  # type:ignore[no-any-return]

    async def create_json(
        self,
        *,
        destination: str,
        cron: str,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        timeout: Optional[Union[str, int]] = None,
        schedule_id: Optional[str] = None,
        queue: Optional[str] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> str:
        """
        Creates a schedule to send messages periodically, automatically serializing the
        body as JSON string, and setting content type to `application/json`.

        Returns the created schedule id.

        :param destination: The destination url or url group.
        :param cron: The cron expression to use to schedule the messages.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param schedule_id: Schedule id to use. This can be used to update the settings
            of an existing schedule.
        :param queue: Name of the queue which the scheduled messages will be enqueued.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        """
        return await self.create(
            destination=destination,
            cron=cron,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            timeout=timeout,
            schedule_id=schedule_id,
            queue=queue,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

    async def get(self, schedule_id: str) -> Schedule:
        """
        Gets the schedule by its id.
        """
        response = await self._http.request(
            path=f"/v2/schedules/{schedule_id}",
            method="GET",
        )

        return parse_schedule_response(response)

    async def list(self) -> List[Schedule]:
        """
        Lists all the schedules.
        """
        response = await self._http.request(
            path="/v2/schedules",
            method="GET",
        )

        return [parse_schedule_response(r) for r in response]

    async def delete(self, schedule_id: str) -> None:
        """
        Deletes the schedule.
        """
        await self._http.request(
            path=f"/v2/schedules/{schedule_id}",
            method="DELETE",
            parse_response=False,
        )

    async def pause(self, schedule_id: str) -> None:
        """
        Pauses the schedule.

        A paused schedule will not produce new messages until
        it is resumed.
        """
        await self._http.request(
            path=f"/v2/schedules/{schedule_id}/pause",
            method="PATCH",
            parse_response=False,
        )

    async def resume(self, schedule_id: str) -> None:
        """
        Resumes the schedule.
        """
        await self._http.request(
            path=f"/v2/schedules/{schedule_id}/resume",
            method="PATCH",
            parse_response=False,
        )



from typing import List

from qstash.asyncio.http import AsyncHttpClient
from qstash.queue import Queue, parse_queue_response, prepare_upsert_body


class AsyncQueueApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def upsert(
        self,
        queue: str,
        *,
        parallelism: int = 1,
        paused: bool = False,
    ) -> None:
        """
        Updates or creates a queue.

        :param queue: The name of the queue.
        :param parallelism: The number of parallel consumers consuming from the queue.
        :param paused: Whether to pause the queue or not. A paused queue will not
            deliver new messages until it is resumed.
        """
        body = prepare_upsert_body(queue, parallelism, paused)

        await self._http.request(
            path="/v2/queues",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    async def get(self, queue: str) -> Queue:
        """
        Gets the queue by its name.
        """
        response = await self._http.request(
            path=f"/v2/queues/{queue}",
            method="GET",
        )

        return parse_queue_response(response)

    async def list(self) -> List[Queue]:
        """
        Lists all the queues.
        """
        response = await self._http.request(
            path="/v2/queues",
            method="GET",
        )

        return [parse_queue_response(r) for r in response]

    async def delete(self, queue: str) -> None:
        """
        Deletes the queue.
        """
        await self._http.request(
            path=f"/v2/queues/{queue}",
            method="DELETE",
            parse_response=False,
        )

    async def pause(self, queue: str) -> None:
        """
        Pauses the queue.

        A paused queue will not deliver messages until
        it is resumed.
        """
        await self._http.request(
            path=f"/v2/queues/{queue}/pause",
            method="POST",
            parse_response=False,
        )

    async def resume(self, queue: str) -> None:
        """
        Resumes the queue.
        """
        await self._http.request(
            path=f"/v2/queues/{queue}/resume",
            method="POST",
            parse_response=False,
        )



import asyncio
from typing import Any, Dict, Literal, Optional, Union

import httpx

from qstash.http import (
    BASE_URL,
    DEFAULT_RETRY,
    NO_RETRY,
    HttpMethod,
    RetryConfig,
    raise_for_non_ok_status,
    DEFAULT_TIMEOUT,
)


class AsyncHttpClient:
    def __init__(
        self,
        token: str,
        retry: Optional[Union[Literal[False], RetryConfig]],
        base_url: Optional[str] = None,
    ) -> None:
        self._token = f"Bearer {token}"

        if retry is None:
            self._retry = DEFAULT_RETRY
        elif retry is False:
            self._retry = NO_RETRY
        else:
            self._retry = retry

        self._client = httpx.AsyncClient(
            timeout=DEFAULT_TIMEOUT,
        )

        self._base_url = base_url.rstrip("/") if base_url else BASE_URL

    async def request(
        self,
        *,
        path: str,
        method: HttpMethod,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Union[str, bytes]] = None,
        params: Optional[Dict[str, str]] = None,
        parse_response: bool = True,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
    ) -> Any:
        base_url = base_url or self._base_url
        token = token or self._token

        url = base_url + path
        headers = {"Authorization": token, **(headers or {})}

        max_attempts = 1 + max(0, self._retry["retries"])
        last_error = None
        response = None
        for attempt in range(max_attempts):
            try:
                response = await self._client.request(
                    method=method,
                    url=url,
                    params=params,
                    headers=headers,
                    content=body,
                )
                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e
                backoff = self._retry["backoff"](attempt) / 1000
                await asyncio.sleep(backoff)

        if not response:
            # Can't be None at this point
            raise last_error  # type:ignore[misc]

        raise_for_non_ok_status(response)

        if parse_response:
            return response.json()

        return response.text

    async def stream(
        self,
        *,
        path: str,
        method: HttpMethod,
        headers: Optional[Dict[str, str]] = None,
        body: Optional[Union[str, bytes]] = None,
        params: Optional[Dict[str, str]] = None,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
    ) -> httpx.Response:
        base_url = base_url or self._base_url
        token = token or self._token

        url = base_url + path
        headers = {"Authorization": token, **(headers or {})}

        max_attempts = 1 + max(0, self._retry["retries"])
        last_error = None
        response = None
        for attempt in range(max_attempts):
            try:
                request = self._client.build_request(
                    method=method,
                    url=url,
                    params=params,
                    headers=headers,
                    content=body,
                )
                response = await self._client.send(
                    request,
                    stream=True,
                )
                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e
                backoff = self._retry["backoff"](attempt) / 1000
                await asyncio.sleep(backoff)

        if not response:
            # Can't be None at this point
            raise last_error  # type:ignore[misc]

        try:
            raise_for_non_ok_status(response)
        except Exception as e:
            await response.aclose()
            raise e

        return response



from os import environ
from typing import Literal, Optional, Union

from qstash.asyncio.dlq import AsyncDlqApi
from qstash.asyncio.flow_control import AsyncFlowControlApi
from qstash.asyncio.log import AsyncLogApi
from qstash.asyncio.http import AsyncHttpClient
from qstash.asyncio.message import AsyncMessageApi
from qstash.asyncio.queue import AsyncQueueApi
from qstash.asyncio.schedule import AsyncScheduleApi
from qstash.asyncio.signing_key import AsyncSigningKeyApi
from qstash.asyncio.url_group import AsyncUrlGroupApi
from qstash.http import RetryConfig


class AsyncQStash:
    def __init__(
        self,
        token: str,
        *,
        retry: Optional[Union[Literal[False], RetryConfig]] = None,
        base_url: Optional[str] = None,
    ) -> None:
        """
        :param token: The authorization token from the Upstash console.
        :param retry: Configures how the client should retry requests.
        """
        self.http = AsyncHttpClient(
            token,
            retry,
            base_url or environ.get("QSTASH_URL"),
        )
        self.message = AsyncMessageApi(self.http)
        """Message api."""

        self.url_group = AsyncUrlGroupApi(self.http)
        """Url group api."""

        self.queue = AsyncQueueApi(self.http)
        """Queue api."""

        self.schedule = AsyncScheduleApi(self.http)
        """Schedule api."""

        self.signing_key = AsyncSigningKeyApi(self.http)
        """Signing key api."""

        self.log = AsyncLogApi(self.http)
        """Log api."""

        self.dlq = AsyncDlqApi(self.http)
        """Dlq (Dead Letter Queue) api."""

        self.flow_control = AsyncFlowControlApi(self.http)
        """Flow control api."""



from typing import Optional

from qstash.asyncio.http import AsyncHttpClient
from qstash.log import (
    LogFilter,
    ListLogsResponse,
    parse_logs_response,
    prepare_list_logs_request_params,
)


class AsyncLogApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def list(
        self,
        *,
        cursor: Optional[str] = None,
        count: Optional[int] = None,
        filter: Optional[LogFilter] = None,
    ) -> ListLogsResponse:
        """
        Lists all logs that happened, such as message creation or delivery.

        :param cursor: Optional cursor to start listing logs from.
        :param count: The maximum number of logs to return.
            Default and max is `1000`.
        :param filter: Filter to use.
        """
        params = prepare_list_logs_request_params(
            cursor=cursor,
            count=count,
            filter=filter,
        )

        response = await self._http.request(
            path="/v2/events",
            method="GET",
            params=params,
        )

        logs = parse_logs_response(response["events"])

        return ListLogsResponse(
            cursor=response.get("cursor"),
            logs=logs,
        )



import base64
import hashlib
from typing import Optional

import jwt

from qstash.errors import SignatureError


def verify_with_key(
    key: str,
    *,
    signature: str,
    body: str,
    url: Optional[str] = None,
    clock_tolerance: int = 0,
) -> None:
    try:
        decoded = jwt.decode(
            signature,
            key,
            algorithms=["HS256"],
            issuer="Upstash",
            options={
                "require": ["iss", "sub", "exp", "nbf"],
                "leeway": clock_tolerance,
            },
        )
    except jwt.ExpiredSignatureError:
        raise SignatureError("Signature has expired")
    except Exception as e:
        raise SignatureError(f"Error while decoding signature: {e}")

    if url is not None and decoded["sub"] != url:
        raise SignatureError(f"Invalid subject: {decoded['sub']}, want: {url}")

    body_hash = hashlib.sha256(body.encode()).digest()
    body_hash_b64 = base64.urlsafe_b64encode(body_hash).decode().rstrip("=")

    if decoded["body"].rstrip("=") != body_hash_b64:
        raise SignatureError(
            f"Invalid body hash: {decoded['body']}, want: {body_hash_b64}"
        )


class Receiver:
    """Receiver offers a simple way to verify the signature of a request."""

    def __init__(self, current_signing_key: str, next_signing_key: str) -> None:
        """
        :param current_signing_key: The current signing key.
            Get it from `https://console.upstash.com/qstash
        :param next_signing_key: The next signing key.
            Get it from `https://console.upstash.com/qstash
        """
        self._current_signing_key = current_signing_key
        self._next_signing_key = next_signing_key

    def verify(
        self,
        *,
        signature: str,
        body: str,
        url: Optional[str] = None,
        clock_tolerance: int = 0,
    ) -> None:
        """
        Verifies the signature of a request.

        Tries to verify the signature with the current signing key.
        If that fails, maybe because you have rotated the keys recently, it will
        try to verify the signature with the next signing key.

        If that fails, the signature is invalid and a `SignatureError` is thrown.

        :param signature: The signature from the `Upstash-Signature` header.
        :param body: The raw request body.
        :param url: Url of the endpoint where the request was sent to.
            When set to `None`, url is not check.
        :param clock_tolerance: Number of seconds to tolerate when checking
            `nbf` and `exp` claims, to deal with small clock differences
            among different servers.
        """
        try:
            verify_with_key(
                self._current_signing_key,
                signature=signature,
                body=body,
                url=url,
                clock_tolerance=clock_tolerance,
            )
        except SignatureError:
            verify_with_key(
                self._next_signing_key,
                signature=signature,
                body=body,
                url=url,
                clock_tolerance=clock_tolerance,
            )



import dataclasses
import enum
import json
from typing import Any, Dict, List, Optional, Union

from qstash.http import HttpClient, HttpMethod
from qstash.message import (
    FlowControl,
    parse_flow_control,
    FlowControlProperties,
    Redact,
)


class ScheduleState(enum.Enum):
    IN_PROGRESS = "IN_PROGRESS"
    SUCCESS = "SUCCESS"
    FAIL = "FAIL"


@dataclasses.dataclass
class Schedule:
    schedule_id: str
    """Id of the schedule."""

    destination: str
    """Destination url or url group."""

    cron: str
    """Cron expression used to schedule the messages."""

    created_at: int
    """Unix time in milliseconds when the schedule was created."""

    body: Optional[str]
    """Body of the scheduled message if it is composed of UTF-8 characters only."""

    body_base64: Optional[str]
    """Base64 encoded body if the scheduled message body contains non-UTF-8 characters."""

    method: HttpMethod
    """HTTP method to use to deliver the message."""

    headers: Optional[Dict[str, List[str]]]
    """Headers that will be forwarded to destination."""

    callback_headers: Optional[Dict[str, List[str]]]
    """Headers that will be forwarded to callback url."""

    failure_callback_headers: Optional[Dict[str, List[str]]]
    """Headers that will be forwarded to failure callback url."""

    retries: int
    """Number of retries that should be attempted in case of delivery failure."""

    retry_delay_expression: Optional[str]
    """
    The retry delay expression for this DLQ message,
    if retry_delay was set when publishing the message.
    """

    callback: Optional[str]
    """Url which is called each time the message is attempted to be delivered."""

    failure_callback: Optional[str]
    """Url which is called after the message is failed."""

    queue: Optional[str]
    """
    Name of the queue which the messages will be enqueued, 
    if the destination is a queue.
    """

    delay: Optional[int]
    """Delay in seconds before the message is delivered."""

    timeout: Optional[int]
    """HTTP timeout value to use while calling the destination url."""

    caller_ip: Optional[str]
    """IP address of the creator of this schedule."""

    paused: bool
    """Whether the schedule is paused or not."""

    flow_control: Optional[FlowControlProperties]
    """Flow control properties"""

    last_schedule_time: Optional[int]
    """Unix time of the last schedule, in milliseconds."""

    next_schedule_time: Optional[int]
    """Unix time of the next schedule, in milliseconds."""

    last_schedule_states: Optional[Dict[str, ScheduleState]]
    """
    Map of the message ids to schedule states for the 
    published/enqueued messages in the last schedule run.
    """

    error: Optional[str]
    """Error message of the last schedule trigger."""

    label: Optional[str]
    """Label assigned to the schedule for filtering logs."""


def prepare_schedule_headers(
    *,
    cron: str,
    content_type: Optional[str],
    method: Optional[HttpMethod],
    headers: Optional[Dict[str, str]],
    callback_headers: Optional[Dict[str, str]],
    failure_callback_headers: Optional[Dict[str, str]],
    retries: Optional[int],
    retry_delay: Optional[str],
    callback: Optional[str],
    failure_callback: Optional[str],
    delay: Optional[Union[str, int]],
    timeout: Optional[Union[str, int]],
    schedule_id: Optional[str],
    queue: Optional[str],
    flow_control: Optional[FlowControl],
    label: Optional[str],
    redact: Optional[Redact],
) -> Dict[str, str]:
    h = {
        "Upstash-Cron": cron,
    }

    if content_type is not None:
        h["Content-Type"] = content_type

    if method is not None:
        h["Upstash-Method"] = method

    if headers:
        for k, v in headers.items():
            if not k.lower().startswith("upstash-forward-"):
                k = f"Upstash-Forward-{k}"

            h[k] = str(v)

    if callback_headers:
        for k, v in callback_headers.items():
            if not k.lower().startswith("upstash-callback-"):
                k = f"Upstash-Callback-{k}"

            h[k] = str(v)

    if failure_callback_headers:
        for k, v in failure_callback_headers.items():
            if not k.lower().startswith("upstash-failure-callback-"):
                k = f"Upstash-Failure-Callback-{k}"

            h[k] = str(v)

    if retries is not None:
        h["Upstash-Retries"] = str(retries)

    if retry_delay is not None:
        h["Upstash-Retry-Delay"] = retry_delay

    if callback is not None:
        h["Upstash-Callback"] = callback

    if failure_callback is not None:
        h["Upstash-Failure-Callback"] = failure_callback

    if delay is not None:
        if isinstance(delay, int):
            h["Upstash-Delay"] = f"{delay}s"
        else:
            h["Upstash-Delay"] = delay

    if timeout is not None:
        if isinstance(timeout, int):
            h["Upstash-Timeout"] = f"{timeout}s"
        else:
            h["Upstash-Timeout"] = timeout

    if schedule_id is not None:
        h["Upstash-Schedule-Id"] = schedule_id

    if flow_control and "key" in flow_control:
        control_values = []

        if "parallelism" in flow_control:
            control_values.append(f"parallelism={flow_control['parallelism']}")

        if "rate" in flow_control:
            control_values.append(f"rate={flow_control['rate']}")

        if "period" in flow_control:
            period = flow_control["period"]
            if isinstance(period, int):
                period = f"{period}s"

            control_values.append(f"period={period}")

        h["Upstash-Flow-Control-Key"] = flow_control["key"]
        h["Upstash-Flow-Control-Value"] = ", ".join(control_values)

    if queue is not None:
        h["Upstash-Queue-Name"] = queue

    if label is not None:
        h["Upstash-Label"] = label

    if redact is not None:
        redact_parts = []

        if redact.get("body"):
            redact_parts.append("body")

        header_redact = redact.get("header")
        if header_redact is not None:
            if header_redact is True:
                redact_parts.append("header")
            elif isinstance(header_redact, list) and len(header_redact) > 0:
                for header_name in header_redact:
                    redact_parts.append(f"header[{header_name}]")

        if redact_parts:
            h["Upstash-Redact-Fields"] = ",".join(redact_parts)

    return h


def parse_schedule_response(response: Dict[str, Any]) -> Schedule:
    flow_control = parse_flow_control(response)

    if "lastScheduleStates" in response:
        last_states = response["lastScheduleStates"]
        for k, v in last_states.items():
            last_states[k] = ScheduleState(v)
    else:
        last_states = None

    return Schedule(
        schedule_id=response["scheduleId"],
        destination=response["destination"],
        cron=response["cron"],
        created_at=response["createdAt"],
        body=response.get("body"),
        body_base64=response.get("bodyBase64"),
        method=response["method"],
        headers=response.get("header"),
        callback_headers=response.get("callbackHeader"),
        failure_callback_headers=response.get("failureCallbackHeader"),
        retries=response["retries"],
        callback=response.get("callback"),
        failure_callback=response.get("failureCallback"),
        delay=response.get("delay"),
        timeout=response.get("timeout"),
        caller_ip=response.get("callerIP"),
        paused=response.get("isPaused", False),
        queue=response.get("queueName"),
        flow_control=flow_control,
        last_schedule_time=response.get("lastScheduleTime"),
        next_schedule_time=response.get("nextScheduleTime"),
        last_schedule_states=last_states,
        error=response.get("error"),
        retry_delay_expression=response.get("retryDelayExpression"),
        label=response.get("label"),
    )


class ScheduleApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def create(
        self,
        *,
        destination: str,
        cron: str,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        timeout: Optional[Union[str, int]] = None,
        schedule_id: Optional[str] = None,
        queue: Optional[str] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> str:
        """
        Creates a schedule to send messages periodically.

        Returns the created schedule id.

        :param destination: The destination url or url group.
        :param cron: The cron expression to use to schedule the messages.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param schedule_id: Schedule id to use. This can be used to update the settings
            of an existing schedule.
        :param queue: Name of the queue which the scheduled messages will be enqueued.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        req_headers = prepare_schedule_headers(
            cron=cron,
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            timeout=timeout,
            schedule_id=schedule_id,
            queue=queue,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

        response = self._http.request(
            path=f"/v2/schedules/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return response["scheduleId"]  # type:ignore[no-any-return]

    def create_json(
        self,
        *,
        destination: str,
        cron: str,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        timeout: Optional[Union[str, int]] = None,
        schedule_id: Optional[str] = None,
        queue: Optional[str] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> str:
        """
        Creates a schedule to send messages periodically, automatically serializing the
        body as JSON string, and setting content type to `application/json`.

        Returns the created schedule id.

        :param destination: The destination url or url group.
        :param cron: The cron expression to use to schedule the messages.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param schedule_id: Schedule id to use. This can be used to update the settings
            of an existing schedule.
        :param queue: Name of the queue which the scheduled messages will be enqueued.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        return self.create(
            destination=destination,
            cron=cron,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            timeout=timeout,
            schedule_id=schedule_id,
            queue=queue,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

    def get(self, schedule_id: str) -> Schedule:
        """
        Gets the schedule by its id.
        """
        response = self._http.request(
            path=f"/v2/schedules/{schedule_id}",
            method="GET",
        )

        return parse_schedule_response(response)

    def list(self) -> List[Schedule]:
        """
        Lists all the schedules.
        """
        response = self._http.request(
            path="/v2/schedules",
            method="GET",
        )

        return [parse_schedule_response(r) for r in response]

    def delete(self, schedule_id: str) -> None:
        """
        Deletes the schedule.
        """
        self._http.request(
            path=f"/v2/schedules/{schedule_id}",
            method="DELETE",
            parse_response=False,
        )

    def pause(self, schedule_id: str) -> None:
        """
        Pauses the schedule.

        A paused schedule will not produce new messages until
        it is resumed.
        """
        self._http.request(
            path=f"/v2/schedules/{schedule_id}/pause",
            method="PATCH",
            parse_response=False,
        )

    def resume(self, schedule_id: str) -> None:
        """
        Resumes the schedule.
        """
        self._http.request(
            path=f"/v2/schedules/{schedule_id}/resume",
            method="PATCH",
            parse_response=False,
        )



import dataclasses
import json
from typing import Any, Dict, List, Optional, TypedDict

from qstash.errors import QStashError
from qstash.http import HttpClient


class UpsertEndpointRequest(TypedDict, total=False):
    url: str
    """Url of the endpoint"""

    name: str
    """Optional name of the endpoint"""


class RemoveEndpointRequest(TypedDict, total=False):
    url: str
    """Url of the endpoint"""

    name: str
    """Name of the endpoint"""


@dataclasses.dataclass
class Endpoint:
    url: str
    """Url of the endpoint"""

    name: Optional[str]
    """Name of the endpoint"""


@dataclasses.dataclass
class UrlGroup:
    name: str
    """Name of the url group."""

    created_at: int
    """Creation time of the url group, in Unix milliseconds."""

    updated_at: int
    """Last update time of the url group, in Unix milliseconds."""

    endpoints: List[Endpoint]
    """List of endpoints."""


def prepare_add_endpoints_body(
    endpoints: List[UpsertEndpointRequest],
) -> str:
    for e in endpoints:
        if "url" not in e:
            raise QStashError("`url` of the endpoint must be provided.")

    return json.dumps(
        {
            "endpoints": endpoints,
        }
    )


def prepare_remove_endpoints_body(
    endpoints: List[RemoveEndpointRequest],
) -> str:
    for e in endpoints:
        if "url" not in e and "name" not in e:
            raise QStashError(
                "One of `url` or `name` of the endpoint must be provided."
            )

    return json.dumps(
        {
            "endpoints": endpoints,
        }
    )


def parse_url_group_response(response: Dict[str, Any]) -> UrlGroup:
    endpoints = []
    for e in response["endpoints"]:
        endpoints.append(
            Endpoint(
                url=e["url"],
                name=e.get("name"),
            )
        )

    return UrlGroup(
        name=response["name"],
        created_at=response["createdAt"],
        updated_at=response["updatedAt"],
        endpoints=endpoints,
    )


class UrlGroupApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def upsert_endpoints(
        self,
        url_group: str,
        endpoints: List[UpsertEndpointRequest],
    ) -> None:
        """
        Add or updates an endpoint to a url group.

        If the url group or the endpoint does not exist, it will be created.
        If the endpoint exists, it will be updated.
        """
        body = prepare_add_endpoints_body(endpoints)

        self._http.request(
            path=f"/v2/topics/{url_group}/endpoints",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    def remove_endpoints(
        self,
        url_group: str,
        endpoints: List[RemoveEndpointRequest],
    ) -> None:
        """
        Remove one or more endpoints from a url group.

        If all endpoints have been removed, the url group will be deleted.
        """
        body = prepare_remove_endpoints_body(endpoints)

        self._http.request(
            path=f"/v2/topics/{url_group}/endpoints",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    def get(self, url_group: str) -> UrlGroup:
        """
        Gets the url group by its name.
        """
        response = self._http.request(
            path=f"/v2/topics/{url_group}",
            method="GET",
        )

        return parse_url_group_response(response)

    def list(self) -> List[UrlGroup]:
        """
        Lists all the url groups.
        """
        response = self._http.request(
            path="/v2/topics",
            method="GET",
        )

        return [parse_url_group_response(r) for r in response]

    def delete(self, url_group: str) -> None:
        """
        Deletes the url group and all its endpoints.
        """
        self._http.request(
            path=f"/v2/topics/{url_group}",
            method="DELETE",
            parse_response=False,
        )



from os import environ
from typing import Optional, Union, Literal

from qstash.dlq import DlqApi
from qstash.flow_control_api import FlowControlApi
from qstash.log import LogApi
from qstash.http import RetryConfig, HttpClient
from qstash.message import MessageApi
from qstash.queue import QueueApi
from qstash.schedule import ScheduleApi
from qstash.signing_key import SigningKeyApi
from qstash.url_group import UrlGroupApi


class QStash:
    """Synchronous SDK for the Upstash QStash."""

    def __init__(
        self,
        token: str,
        *,
        retry: Optional[Union[Literal[False], RetryConfig]] = None,
        base_url: Optional[str] = None,
    ) -> None:
        """
        :param token: The authorization token from the Upstash console.
        :param retry: Configures how the client should retry requests.
        """
        self.http = HttpClient(
            token,
            retry,
            base_url or environ.get("QSTASH_URL"),
        )
        self.message = MessageApi(self.http)
        """Message api."""

        self.url_group = UrlGroupApi(self.http)
        """Url group api."""

        self.queue = QueueApi(self.http)
        """Queue api."""

        self.schedule = ScheduleApi(self.http)
        """Schedule api."""

        self.signing_key = SigningKeyApi(self.http)
        """Signing key api."""

        self.log = LogApi(self.http)
        """Log api."""

        self.dlq = DlqApi(self.http)
        """Dlq (Dead Letter Queue) api."""

        self.flow_control = FlowControlApi(self.http)
        """Flow control api."""



import dataclasses
import json
from typing import (
    Union,
    Optional,
    Literal,
    Dict,
    Any,
    List,
    TypedDict,
)

from qstash.chat import LlmProvider
from qstash.errors import QStashError
from qstash.http import HttpClient, HttpMethod


class LlmApi(TypedDict):
    name: Literal["llm"]
    """Name of the API type."""

    provider: LlmProvider
    """LLM provider for the API."""


ApiT = LlmApi  # In the future, this can be union of different API types


class FlowControl(TypedDict, total=False):
    key: str
    """Flow control key"""

    parallelism: int
    """Number of requests which can be active with the same key."""

    rate: int
    """Number of requests to activate per period with the same key."""

    period: Union[str, int]
    """
    Unit duration of the rate.
    When given as an integer, it is in seconds. Otherwise, it
    can be specified as a duration string like '10s'(10 seconds),
    '2m'(2 minutes), '1h'(1 hour),
    or '3d5h12m'(3 days and 5 hours and 12 minutes).
    Can be at most a week and defaults to 1 second.
    """


class Redact(TypedDict, total=False):
    body: Literal[True]
    """Redact the request body in logs."""

    header: Union[Literal[True], List[str]]
    """
    Redact headers in logs.
    - `True` to redact all headers
    - List of header names to redact specific headers (e.g., ["Authorization"])
    """


@dataclasses.dataclass
class FlowControlProperties:
    key: str
    """Flow control key."""

    parallelism: Optional[int]
    """Number of requests which can be active with the same key."""

    rate: Optional[int]
    """Number of requests to activate per period with the same key."""

    period: Optional[int]
    """Unit duration of the rate in seconds."""


@dataclasses.dataclass
class PublishResponse:
    message_id: str
    """Unique id of the message."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


@dataclasses.dataclass
class PublishUrlGroupResponse:
    message_id: str
    """Unique id of the message."""

    url: str
    """Url where the message was sent to."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


@dataclasses.dataclass
class EnqueueResponse:
    message_id: str
    """Unique id of the message."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


@dataclasses.dataclass
class EnqueueUrlGroupResponse:
    message_id: str
    """Unique id of the message."""

    url: str
    """Url where the message was sent to."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


@dataclasses.dataclass
class BatchResponse:
    message_id: str
    """Unique id of the message."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


@dataclasses.dataclass
class BatchUrlGroupResponse:
    message_id: str
    """Unique id of the message."""

    url: str
    """Url where the message was sent to."""

    deduplicated: bool
    """Whether the message is a duplicate and was not sent to the destination."""


class BatchRequest(TypedDict, total=False):
    url: str
    """Url to send the message to."""

    url_group: str
    """Url group to send the message to."""

    queue: str
    """Name of the queue that message will be enqueued on."""

    api: ApiT
    """Api to send the message to."""

    body: Union[str, bytes]
    """Raw request message body passed to the endpoints as is."""

    content_type: str
    """MIME type of the message."""

    method: HttpMethod
    """HTTP method to use when sending a webhook to your API."""

    headers: Dict[str, str]
    """Headers to forward along with the message."""

    callback_headers: Dict[str, str]
    """Headers to forward along with the callback message."""

    failure_callback_headers: Dict[str, str]
    """Headers to forward along with the failure callback message."""

    retries: int
    """
    How often should this message be retried in case the destination 
    API is not available.
    """

    retry_delay: str
    """
    Delay between retries.

    By default, the `retryDelay` is exponential backoff.
    More details can be found in: https://upstash.com/docs/qstash/features/retry.
    
    The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.
    
    You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
    The special variable `retried` represents the current retry attempt count (starting from 0).
    
    Supported functions:
    - `pow`
    - `sqrt`
    - `abs`
    - `exp`
    - `floor`
    - `ceil`
    - `round`
    - `min`
    - `max`
    
    Examples of valid `retryDelay` values:
    ```py
    1000 # 1 second
    1000 * (1 + retried) # 1 second multiplied by the current retry attempt
    pow(2, retried) # 2 to the power of the current retry attempt
    max(10, pow(2, retried)) # The greater of 10 or 2^retried
    ```
    """

    callback: str
    """A callback url that will be called after each attempt."""

    failure_callback: str
    """
    A failure callback url that will be called when a delivery is failed, 
    that is when all the defined retries are exhausted.
    """

    delay: Union[str, int]
    """
    Delay the message delivery. 
    
    The format for the delay string is a
    number followed by duration abbreviation, like `10s`. Available durations
    are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
    it is also possible to specify the delay as an integer, which will be
    interpreted as delay in seconds.
    """

    not_before: int
    """
    Delay the message until a certain time in the future. 
    
    The format is a unix timestamp in seconds, based on the UTC timezone.
    """

    deduplication_id: str
    """Id to use while deduplicating messages."""

    content_based_deduplication: bool
    """Automatically deduplicate messages based on their content."""

    timeout: Union[str, int]
    """
    The HTTP timeout value to use while calling the destination URL.
    When a timeout is specified, it will be used instead of the maximum timeout
    value permitted by the QStash plan. It is useful in scenarios, where a message
    should be delivered with a shorter timeout.
    
    The format for the timeout string is a number followed by duration abbreviation, 
    like `10s`. Available durations are `s` (seconds), `m` (minutes), `h` (hours), 
    and `d` (days). As convenience, it is also possible to specify the timeout as 
    an integer, which will be interpreted as timeout in seconds.
    """

    flow_control: FlowControl
    """
    Settings for controlling the number of active requests, as well as
    the rate of requests with the same flow control key.
    """

    label: str
    """Assign a label to the request to filter logs with it later."""

    redact: Redact
    """
    Configure which fields should be redacted in logs.
    - `{"body": True}` to redact the request body
    - `{"header": True}` to redact all headers
    - `{"header": ["Authorization"]}` to redact specific headers
    """


class BatchJsonRequest(TypedDict, total=False):
    url: str
    """Url to send the message to."""

    url_group: str
    """Url group to send the message to."""

    queue: str
    """Name of the queue that message will be enqueued on."""

    api: ApiT
    """Api to send the message to."""

    body: Any
    """
    The request body passed to the endpoints after being serialized to 
    JSON string.
    """

    method: HttpMethod
    """HTTP method to use when sending a webhook to your API."""

    headers: Dict[str, str]
    """Headers to forward along with the message."""

    callback_headers: Dict[str, str]
    """Headers to forward along with the callback message."""

    failure_callback_headers: Dict[str, str]
    """Headers to forward along with the failure callback message."""

    retries: int
    """
    How often should this message be retried in case the destination 
    API is not available.
    """

    retry_delay: str
    """
    Delay between retries.

    By default, the `retryDelay` is exponential backoff.
    More details can be found in: https://upstash.com/docs/qstash/features/retry.
    
    The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.
    
    You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
    The special variable `retried` represents the current retry attempt count (starting from 0).
    
    Supported functions:
    - `pow`
    - `sqrt`
    - `abs`
    - `exp`
    - `floor`
    - `ceil`
    - `round`
    - `min`
    - `max`
    
    Examples of valid `retryDelay` values:
    ```py
    1000 # 1 second
    1000 * (1 + retried) # 1 second multiplied by the current retry attempt
    pow(2, retried) # 2 to the power of the current retry attempt
    max(10, pow(2, retried)) # The greater of 10 or 2^retried
    ```
    """

    callback: str
    """A callback url that will be called after each attempt."""

    failure_callback: str
    """
    A failure callback url that will be called when a delivery is failed, 
    that is when all the defined retries are exhausted.
    """

    delay: Union[str, int]
    """
    Delay the message delivery. 
    
    The format for the delay string is a
    number followed by duration abbreviation, like `10s`. Available durations
    are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
    it is also possible to specify the delay as an integer, which will be
    interpreted as delay in seconds.
    """

    not_before: int
    """
    Delay the message until a certain time in the future. 
    
    The format is a unix timestamp in seconds, based on the UTC timezone.
    """

    deduplication_id: str
    """Id to use while deduplicating messages."""

    content_based_deduplication: bool
    """Automatically deduplicate messages based on their content."""

    timeout: Union[str, int]
    """
    The HTTP timeout value to use while calling the destination URL.
    When a timeout is specified, it will be used instead of the maximum timeout
    value permitted by the QStash plan. It is useful in scenarios, where a message
    should be delivered with a shorter timeout.

    The format for the timeout string is a number followed by duration abbreviation, 
    like `10s`. Available durations are `s` (seconds), `m` (minutes), `h` (hours), 
    and `d` (days). As convenience, it is also possible to specify the timeout as 
    an integer, which will be interpreted as timeout in seconds.
    """

    flow_control: FlowControl
    """
    Settings for controlling the number of active requests, as well as
    the rate of requests with the same flow control key.
    """

    label: str
    """Assign a label to the request to filter logs with it later."""

    redact: Redact
    """
    Configure which fields should be redacted in logs.
    - `{"body": True}` to redact the request body
    - `{"header": True}` to redact all headers
    - `{"header": ["Authorization"]}` to redact specific headers
    """


@dataclasses.dataclass
class Message:
    message_id: str
    """Id of the message."""

    url: str
    """Destination url of the message."""

    url_group: Optional[str]
    """Url group name if this message was sent to a url group."""

    endpoint: Optional[str]
    """
    Endpoint name of the message if the endpoint is given a 
    name within the url group.
    """

    queue: Optional[str]
    """Queue name if this message was enqueued to a queue."""

    body: Optional[str]
    """Body of the message if it is composed of UTF-8 characters only."""

    body_base64: Optional[str]
    """Base64 encoded body if the body contains non-UTF-8 characters."""

    method: HttpMethod
    """HTTP method used to deliver the message."""

    headers: Optional[Dict[str, List[str]]]
    """Headers forwarded to the url."""

    callback_headers: Optional[Dict[str, List[str]]]
    """Headers forwarded to callback url."""

    failure_callback_headers: Optional[Dict[str, List[str]]]
    """Headers forwarded to failure callback url."""

    max_retries: int
    """Number of retries that should be attempted in case of delivery failure."""

    retry_delay_expression: Optional[str]
    """
    The retry delay expression for this DLQ message,
    if retry_delay was set when publishing the message.
    """

    not_before: int
    """Unix time in milliseconds before which the message should not be delivered."""

    created_at: int
    """Unix time in milliseconds when the message was created."""

    callback: Optional[str]
    """Url which is called each time the message is attempted to be delivered."""

    failure_callback: Optional[str]
    """Url which is called after the message is failed."""

    schedule_id: Optional[str]
    """Schedule id of the message if the message is triggered by a schedule."""

    caller_ip: Optional[str]
    """IP address of the publisher of this message."""

    flow_control: Optional[FlowControlProperties]
    """Flow control properties."""

    label: Optional[str]
    """Label assigned to the request for filtering logs."""


def get_destination(
    *,
    url: Optional[str],
    url_group: Optional[str],
    api: Optional[ApiT],
    headers: Dict[str, str],
) -> str:
    destination = None
    count = 0
    if url is not None:
        destination = url
        count += 1

    if url_group is not None:
        destination = url_group
        count += 1

    if api is not None:
        provider = api["provider"]
        destination = provider.base_url + "/v1/chat/completions"
        headers["Authorization"] = f"Bearer {provider.token}"

        count += 1

    if count != 1:
        raise QStashError(
            "Only and only one of 'url', 'url_group', or 'api' must be provided."
        )

    # Can't be None at this point
    return destination  # type:ignore[return-value]


def prepare_headers(
    *,
    content_type: Optional[str],
    method: Optional[HttpMethod],
    headers: Optional[Dict[str, str]],
    callback_headers: Optional[Dict[str, str]],
    failure_callback_headers: Optional[Dict[str, str]],
    retries: Optional[int],
    retry_delay: Optional[str],
    callback: Optional[str],
    failure_callback: Optional[str],
    delay: Optional[Union[str, int]],
    not_before: Optional[int],
    deduplication_id: Optional[str],
    content_based_deduplication: Optional[bool],
    timeout: Optional[Union[str, int]],
    flow_control: Optional[FlowControl],
    label: Optional[str],
    redact: Optional[Redact],
) -> Dict[str, str]:
    h = {}

    if content_type is not None:
        h["Content-Type"] = content_type

    if method is not None:
        h["Upstash-Method"] = method

    if headers:
        for k, v in headers.items():
            if not k.lower().startswith("upstash-forward-"):
                k = f"Upstash-Forward-{k}"

            h[k] = str(v)

    if callback_headers:
        for k, v in callback_headers.items():
            if not k.lower().startswith("upstash-callback-"):
                k = f"Upstash-Callback-{k}"

            h[k] = str(v)

    if failure_callback_headers:
        for k, v in failure_callback_headers.items():
            if not k.lower().startswith("upstash-failure-callback-"):
                k = f"Upstash-Failure-Callback-{k}"

            h[k] = str(v)

    if retries is not None:
        h["Upstash-Retries"] = str(retries)

    if retry_delay is not None:
        h["Upstash-Retry-Delay"] = retry_delay

    if callback is not None:
        h["Upstash-Callback"] = callback

    if failure_callback is not None:
        h["Upstash-Failure-Callback"] = failure_callback

    if delay is not None:
        if isinstance(delay, int):
            h["Upstash-Delay"] = f"{delay}s"
        else:
            h["Upstash-Delay"] = delay

    if not_before is not None:
        h["Upstash-Not-Before"] = str(not_before)

    if deduplication_id is not None:
        h["Upstash-Deduplication-Id"] = deduplication_id

    if content_based_deduplication is not None:
        h["Upstash-Content-Based-Deduplication"] = str(content_based_deduplication)

    if timeout is not None:
        if isinstance(timeout, int):
            h["Upstash-Timeout"] = f"{timeout}s"
        else:
            h["Upstash-Timeout"] = timeout

    if flow_control and "key" in flow_control:
        control_values = []

        if "parallelism" in flow_control:
            control_values.append(f"parallelism={flow_control['parallelism']}")

        if "rate" in flow_control:
            control_values.append(f"rate={flow_control['rate']}")

        if "period" in flow_control:
            period = flow_control["period"]
            if isinstance(period, int):
                period = f"{period}s"

            control_values.append(f"period={period}")

        h["Upstash-Flow-Control-Key"] = flow_control["key"]
        h["Upstash-Flow-Control-Value"] = ", ".join(control_values)

    if label is not None:
        h["Upstash-Label"] = label

    if redact is not None:
        redact_parts = []

        if redact.get("body"):
            redact_parts.append("body")

        header_redact = redact.get("header")
        if header_redact is not None:
            if header_redact is True:
                redact_parts.append("header")
            elif isinstance(header_redact, list) and len(header_redact) > 0:
                for header_name in header_redact:
                    redact_parts.append(f"header[{header_name}]")

        if redact_parts:
            h["Upstash-Redact-Fields"] = ",".join(redact_parts)

    return h


def parse_publish_response(
    response: Union[List[Dict[str, Any]], Dict[str, Any]],
) -> Union[PublishResponse, List[PublishUrlGroupResponse]]:
    if isinstance(response, list):
        result = []
        for ug_resp in response:
            result.append(
                PublishUrlGroupResponse(
                    message_id=ug_resp["messageId"],
                    url=ug_resp["url"],
                    deduplicated=ug_resp.get("deduplicated", False),
                )
            )

        return result

    return PublishResponse(
        message_id=response["messageId"],
        deduplicated=response.get("deduplicated", False),
    )


def parse_enqueue_response(
    response: Union[List[Dict[str, Any]], Dict[str, Any]],
) -> Union[EnqueueResponse, List[EnqueueUrlGroupResponse]]:
    if isinstance(response, list):
        result = []
        for ug_resp in response:
            result.append(
                EnqueueUrlGroupResponse(
                    message_id=ug_resp["messageId"],
                    url=ug_resp["url"],
                    deduplicated=ug_resp.get("deduplicated", False),
                )
            )

        return result

    return EnqueueResponse(
        message_id=response["messageId"],
        deduplicated=response.get("deduplicated", False),
    )


def prepare_batch_message_body(messages: List[BatchRequest]) -> str:
    batch_messages = []

    for msg in messages:
        user_headers = msg.get("headers", {})
        destination = get_destination(
            url=msg.get("url"),
            url_group=msg.get("url_group"),
            api=msg.get("api"),
            headers=user_headers,
        )

        headers = prepare_headers(
            content_type=msg.get("content_type"),
            method=msg.get("method"),
            headers=user_headers,
            callback_headers=msg.get("callback_headers"),
            failure_callback_headers=msg.get("failure_callback_headers"),
            retries=msg.get("retries"),
            retry_delay=msg.get("retry_delay"),
            callback=msg.get("callback"),
            failure_callback=msg.get("failure_callback"),
            delay=msg.get("delay"),
            not_before=msg.get("not_before"),
            deduplication_id=msg.get("deduplication_id"),
            content_based_deduplication=msg.get("content_based_deduplication"),
            timeout=msg.get("timeout"),
            flow_control=msg.get("flow_control"),
            label=msg.get("label"),
            redact=msg.get("redact"),
        )

        batch_messages.append(
            {
                "destination": destination,
                "headers": headers,
                "body": msg.get("body"),
                "queue": msg.get("queue"),
            }
        )

    return json.dumps(batch_messages)


def parse_batch_response(
    response: List[Union[List[Dict[str, Any]], Dict[str, Any]]],
) -> List[Union[BatchResponse, List[BatchUrlGroupResponse]]]:
    result: List[Union[BatchResponse, List[BatchUrlGroupResponse]]] = []

    for resp in response:
        if isinstance(resp, list):
            ug_result = []
            for ug_resp in resp:
                ug_result.append(
                    BatchUrlGroupResponse(
                        message_id=ug_resp["messageId"],
                        url=ug_resp["url"],
                        deduplicated=ug_resp.get("deduplicated", False),
                    )
                )

            result.append(ug_result)
        else:
            result.append(
                BatchResponse(
                    message_id=resp["messageId"],
                    deduplicated=resp.get("deduplicated", False),
                )
            )

    return result


def convert_to_batch_messages(
    messages: List[BatchJsonRequest],
) -> List[BatchRequest]:
    batch_messages = []

    for msg in messages:
        batch_msg: BatchRequest = {}
        if "queue" in msg:
            batch_msg["queue"] = msg["queue"]

        if "url" in msg:
            batch_msg["url"] = msg["url"]

        if "url_group" in msg:
            batch_msg["url_group"] = msg["url_group"]

        if "api" in msg:
            batch_msg["api"] = msg["api"]

        batch_msg["body"] = json.dumps(msg.get("body"))
        batch_msg["content_type"] = "application/json"

        if "method" in msg:
            batch_msg["method"] = msg["method"]

        if "headers" in msg:
            batch_msg["headers"] = msg["headers"]

        if "callback_headers" in msg:
            batch_msg["callback_headers"] = msg["callback_headers"]

        if "failure_callback_headers" in msg:
            batch_msg["failure_callback_headers"] = msg["failure_callback_headers"]

        if "retries" in msg:
            batch_msg["retries"] = msg["retries"]

        if "callback" in msg:
            batch_msg["callback"] = msg["callback"]

        if "failure_callback" in msg:
            batch_msg["failure_callback"] = msg["failure_callback"]

        if "delay" in msg:
            batch_msg["delay"] = msg["delay"]

        if "not_before" in msg:
            batch_msg["not_before"] = msg["not_before"]

        if "deduplication_id" in msg:
            batch_msg["deduplication_id"] = msg["deduplication_id"]

        if "content_based_deduplication" in msg:
            batch_msg["content_based_deduplication"] = msg[
                "content_based_deduplication"
            ]

        if "timeout" in msg:
            batch_msg["timeout"] = msg["timeout"]

        if "flow_control" in msg:
            batch_msg["flow_control"] = msg["flow_control"]

        if "label" in msg:
            batch_msg["label"] = msg["label"]

        if "redact" in msg:
            batch_msg["redact"] = msg["redact"]

        batch_messages.append(batch_msg)

    return batch_messages


def parse_flow_control(response: Dict[str, Any]) -> Optional[FlowControlProperties]:
    if "flowControlKey" not in response:
        return None

    return FlowControlProperties(
        key=response["flowControlKey"],
        parallelism=response.get("parallelism"),
        rate=response.get("rate"),
        period=response.get("period"),
    )


def parse_message_response(response: Dict[str, Any]) -> Message:
    flow_control = parse_flow_control(response)
    return Message(
        message_id=response["messageId"],
        url=response["url"],
        url_group=response.get("topicName"),
        endpoint=response.get("endpointName"),
        queue=response.get("queueName"),
        body=response.get("body"),
        body_base64=response.get("bodyBase64"),
        method=response["method"],
        headers=response.get("header"),
        callback_headers=response.get("callbackHeader"),
        failure_callback_headers=response.get("failureCallbackHeader"),
        max_retries=response["maxRetries"],
        not_before=response["notBefore"],
        created_at=response["createdAt"],
        callback=response.get("callback"),
        failure_callback=response.get("failureCallback"),
        schedule_id=response.get("scheduleId"),
        caller_ip=response.get("callerIP"),
        flow_control=flow_control,
        retry_delay_expression=response.get("retryDelayExpression"),
        label=response.get("label"),
    )


class MessageApi:
    def __init__(self, http: HttpClient):
        self._http = http

    def publish(
        self,
        *,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        not_before: Optional[int] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[PublishResponse, List[PublishUrlGroupResponse]]:
        """
        Publishes a message to QStash.

        If the destination is a `url` or an `api`, `PublishResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `PublishUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param not_before: Delay the message until a certain time in the future.
            The format is a unix timestamp in seconds, based on the UTC timezone.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        headers = headers or {}
        destination = get_destination(
            url=url,
            url_group=url_group,
            api=api,
            headers=headers,
        )

        req_headers = prepare_headers(
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            not_before=not_before,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

        response = self._http.request(
            path=f"/v2/publish/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return parse_publish_response(response)

    def publish_json(
        self,
        *,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        delay: Optional[Union[str, int]] = None,
        not_before: Optional[int] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        flow_control: Optional[FlowControl] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[PublishResponse, List[PublishUrlGroupResponse]]:
        """
        Publish a message to QStash, automatically serializing the
        body as JSON string, and setting content type to `application/json`.

        If the destination is a `url` or an `api`, `PublishResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `PublishUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param delay: Delay the message delivery. The format for the delay string is a
            number followed by duration abbreviation, like `10s`. Available durations
            are `s` (seconds), `m` (minutes), `h` (hours), and `d` (days). As convenience,
            it is also possible to specify the delay as an integer, which will be
            interpreted as delay in seconds.
        :param not_before: Delay the message until a certain time in the future.
            The format is a unix timestamp in seconds, based on the UTC timezone.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param flow_control: Settings for controlling the number of active requests,
            as well as the rate of requests with the same flow control key.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        return self.publish(
            url=url,
            url_group=url_group,
            api=api,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=delay,
            not_before=not_before,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=flow_control,
            label=label,
            redact=redact,
        )

    def enqueue(
        self,
        *,
        queue: str,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Union[str, bytes]] = None,
        content_type: Optional[str] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[EnqueueResponse, List[EnqueueUrlGroupResponse]]:
        """
        Enqueues a message, after creating the queue if it does
        not exist.

        If the destination is a `url` or an `api`, `EnqueueResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `EnqueueUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param queue: The name of the queue.
        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The raw request message body passed to the destination as is.
        :param content_type: MIME type of the message.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        headers = headers or {}
        destination = get_destination(
            url=url,
            url_group=url_group,
            api=api,
            headers=headers,
        )

        req_headers = prepare_headers(
            content_type=content_type,
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            delay=None,
            not_before=None,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            flow_control=None,
            label=label,
            redact=redact,
        )

        response = self._http.request(
            path=f"/v2/enqueue/{queue}/{destination}",
            method="POST",
            headers=req_headers,
            body=body,
        )

        return parse_enqueue_response(response)

    def enqueue_json(
        self,
        *,
        queue: str,
        url: Optional[str] = None,
        url_group: Optional[str] = None,
        api: Optional[ApiT] = None,
        body: Optional[Any] = None,
        method: Optional[HttpMethod] = None,
        headers: Optional[Dict[str, str]] = None,
        callback_headers: Optional[Dict[str, str]] = None,
        failure_callback_headers: Optional[Dict[str, str]] = None,
        retries: Optional[int] = None,
        retry_delay: Optional[str] = None,
        callback: Optional[str] = None,
        failure_callback: Optional[str] = None,
        deduplication_id: Optional[str] = None,
        content_based_deduplication: Optional[bool] = None,
        timeout: Optional[Union[str, int]] = None,
        label: Optional[str] = None,
        redact: Optional[Redact] = None,
    ) -> Union[EnqueueResponse, List[EnqueueUrlGroupResponse]]:
        """
        Enqueues a message, after creating the queue if it does
        not exist. It automatically serializes the body as JSON string,
        and setting content type to `application/json`.

        If the destination is a `url` or an `api`, `EnqueueResponse`
        is returned.

        If the destination is a `url_group`, then a list of
        `EnqueueUrlGroupResponse`s are returned, one for each url
        in the url group.

        :param queue: The name of the queue.
        :param url: Url to send the message to.
        :param url_group: Url group to send the message to.
        :param api: Api to send the message to.
        :param body: The request message body passed to the destination after being
            serialized as JSON string.
        :param method: The HTTP method to use when sending a webhook to your API.
        :param headers: Headers to forward along with the message.
        :param callback_headers: Headers to forward along with the callback message.
        :param failure_callback_headers: Headers to forward along with the failure
            callback message.
        :param retries: How often should this message be retried in case the destination
            API is not available.
        :param retry_delay: Delay between retries.

            By default, the `retryDelay` is exponential backoff.
            More details can be found in: https://upstash.com/docs/qstash/features/retry.

            The `retryDelay` option allows you to customize the delay (in milliseconds) between retry attempts when message delivery fails.

            You can use mathematical expressions and the following built-in functions to calculate the delay dynamically.
            The special variable `retried` represents the current retry attempt count (starting from 0).

            Supported functions:
            - `pow`
            - `sqrt`
            - `abs`
            - `exp`
            - `floor`
            - `ceil`
            - `round`
            - `min`
            - `max`

            Examples of valid `retryDelay` values:
            ```py
            1000 # 1 second
            1000 * (1 + retried) # 1 second multiplied by the current retry attempt
            pow(2, retried) # 2 to the power of the current retry attempt
            max(10, pow(2, retried)) # The greater of 10 or 2^retried
            ```
        :param callback: A callback url that will be called after each attempt.
        :param failure_callback: A failure callback url that will be called when a delivery
            is failed, that is when all the defined retries are exhausted.
        :param deduplication_id: Id to use while deduplicating messages.
        :param content_based_deduplication: Automatically deduplicate messages based on
            their content.
        :param timeout: The HTTP timeout value to use while calling the destination URL.
            When a timeout is specified, it will be used instead of the maximum timeout
            value permitted by the QStash plan. It is useful in scenarios, where a message
            should be delivered with a shorter timeout.
        :param label: Assign a label to the request to filter logs with it later.
        :param redact: Configure which fields should be redacted in logs.
            - `{"body": True}` to redact the request body
            - `{"header": True}` to redact all headers
            - `{"header": ["Authorization"]}` to redact specific headers
        """
        return self.enqueue(
            queue=queue,
            url=url,
            url_group=url_group,
            api=api,
            body=json.dumps(body),
            content_type="application/json",
            method=method,
            headers=headers,
            callback_headers=callback_headers,
            failure_callback_headers=failure_callback_headers,
            retries=retries,
            retry_delay=retry_delay,
            callback=callback,
            failure_callback=failure_callback,
            deduplication_id=deduplication_id,
            content_based_deduplication=content_based_deduplication,
            timeout=timeout,
            label=label,
            redact=redact,
        )

    def batch(
        self, messages: List[BatchRequest]
    ) -> List[Union[BatchResponse, List[BatchUrlGroupResponse]]]:
        """
        Publishes or enqueues multiple messages in a single request.

        Returns a list of publish or enqueue responses, one for each
        message in the batch.

        If the message in the batch is sent to a url or an API,
        the corresponding item in the response is `BatchResponse`.

        If the message in the batch is sent to a url group,
        the corresponding item in the response is list of
        `BatchUrlGroupResponse`s, one for each url in the url group.
        """
        body = prepare_batch_message_body(messages)

        response = self._http.request(
            path="/v2/batch",
            body=body,
            headers={"Content-Type": "application/json"},
            method="POST",
        )

        return parse_batch_response(response)

    def batch_json(
        self, messages: List[BatchJsonRequest]
    ) -> List[Union[BatchResponse, List[BatchUrlGroupResponse]]]:
        """
        Publishes or enqueues multiple messages in a single request,
        automatically serializing the message bodies as JSON strings,
        and setting content type to `application/json`.

        Returns a list of publish or enqueue responses, one for each
        message in the batch.

        If the message in the batch is sent to a url or an API,
        the corresponding item in the response is `BatchResponse`.

        If the message in the batch is sent to a url group,
        the corresponding item in the response is list of
        `BatchUrlGroupResponse`s, one for each url in the url group.
        """
        batch_messages = convert_to_batch_messages(messages)
        return self.batch(batch_messages)

    def get(self, message_id: str) -> Message:
        """
        Gets the message by its id.
        """
        response = self._http.request(
            path=f"/v2/messages/{message_id}",
            method="GET",
        )

        return parse_message_response(response)

    def cancel(self, message_id: str) -> None:
        """
        Cancels delivery of an existing message.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.
        """
        self._http.request(
            path=f"/v2/messages/{message_id}",
            method="DELETE",
            parse_response=False,
        )

    def cancel_many(self, message_ids: List[str]) -> int:
        """
        Cancels delivery of existing messages.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.

        Returns how many of the messages are cancelled.
        """
        body = json.dumps({"messageIds": message_ids})

        response = self._http.request(
            path="/v2/messages",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
        )

        return response["cancelled"]  # type:ignore[no-any-return]

    def cancel_all(self) -> int:
        """
        Cancels delivery of all the existing messages.

        Cancelling a message will remove it from QStash and stop it from being
        delivered in the future. If a message is in flight to your API,
        it might be too late to cancel.

        Returns how many messages are cancelled.
        """
        response = self._http.request(
            path="/v2/messages",
            method="DELETE",
        )

        return response["cancelled"]  # type:ignore[no-any-return]



import json
from typing import List, Optional

from qstash.asyncio.http import AsyncHttpClient
from qstash.dlq import (
    DlqMessage,
    ListDlqMessagesResponse,
    parse_dlq_message_response,
    DlqFilter,
    prepare_list_dlq_messages_params,
)


class AsyncDlqApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def get(self, dlq_id: str) -> DlqMessage:
        """
        Gets a message from DLQ.

        :param dlq_id: The unique id within the DLQ to get.
        """
        response = await self._http.request(
            path=f"/v2/dlq/{dlq_id}",
            method="GET",
        )

        return parse_dlq_message_response(response, dlq_id)

    async def list(
        self,
        *,
        cursor: Optional[str] = None,
        count: Optional[int] = None,
        filter: Optional[DlqFilter] = None,
    ) -> ListDlqMessagesResponse:
        """
        Lists all messages currently inside the DLQ.

        :param cursor: Optional cursor to start listing DLQ messages from.
        :param count: The maximum number of DLQ messages to return.
            Default and max is `100`.
        :param filter: Filter to use.
        """
        params = prepare_list_dlq_messages_params(
            cursor=cursor,
            count=count,
            filter=filter,
        )

        response = await self._http.request(
            path="/v2/dlq",
            method="GET",
            params=params,
        )

        messages = [parse_dlq_message_response(r) for r in response["messages"]]

        return ListDlqMessagesResponse(
            cursor=response.get("cursor"),
            messages=messages,
        )

    async def delete(self, dlq_id: str) -> None:
        """
        Deletes a message from the DLQ.

        :param dlq_id: The unique id within the DLQ to delete.
        """
        await self._http.request(
            path=f"/v2/dlq/{dlq_id}",
            method="DELETE",
            parse_response=False,
        )

    async def delete_many(self, dlq_ids: List[str]) -> int:
        """
        Deletes multiple messages from the DLQ and
        returns how many of them are deleted.

        :param dlq_ids: The unique ids within the DLQ to delete.
        """
        body = json.dumps({"dlqIds": dlq_ids})

        response = await self._http.request(
            path="/v2/dlq",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
        )

        return response["deleted"]  # type:ignore[no-any-return]



from typing import Optional


class QStashError(Exception): ...


class SignatureError(QStashError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class RateLimitExceededError(QStashError):
    def __init__(
        self, limit: Optional[str], remaining: Optional[str], reset: Optional[str]
    ):
        super().__init__(
            f"Exceeded burst rate limit: Limit: {limit}, remaining: {remaining}, reset: {reset}"
        )
        self.limit = limit
        self.remaining = remaining
        self.reset = reset


class DailyMessageLimitExceededError(QStashError):
    def __init__(
        self, limit: Optional[str], remaining: Optional[str], reset: Optional[str]
    ):
        super().__init__(
            f"Exceeded daily message limit: Limit: {limit}, remaining: {remaining}, reset: {reset}"
        )
        self.limit = limit
        self.remaining = remaining
        self.reset = reset



import dataclasses
from typing import Any, Dict

from qstash.http import HttpClient


@dataclasses.dataclass
class SigningKey:
    current: str
    """Current signing key."""

    next: str
    """Next signing key."""


def parse_signing_key_response(response: Dict[str, Any]) -> SigningKey:
    return SigningKey(
        current=response["current"],
        next=response["next"],
    )


class SigningKeyApi:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def get(self) -> SigningKey:
        """
        Gets the current and next signing keys.
        """
        response = self._http.request(
            path="/v2/keys",
            method="GET",
        )

        return parse_signing_key_response(response)

    def rotate(self) -> SigningKey:
        """
        Rotates the current signing key and gets the new signing key.

        The next signing key becomes the current signing
        key, and a new signing key is assigned to the
        next signing key.
        """
        response = self._http.request(
            path="/v2/keys/rotate",
            method="POST",
        )

        return parse_signing_key_response(response)



from typing import Dict

from qstash.asyncio.http import AsyncHttpClient
from qstash.flow_control_api import (
    FlowControlInfo,
    GlobalParallelismInfo,
    PinFlowControlOptions,
    UnpinFlowControlOptions,
    parse_flow_control_info,
)


class AsyncFlowControlApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def get(self, flow_control_key: str) -> FlowControlInfo:
        """
        Gets a single flow control by key.

        :param flow_control_key: The flow control key to get.
        """
        response = await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}",
            method="GET",
        )

        return parse_flow_control_info(response)

    async def get_global_parallelism(self) -> GlobalParallelismInfo:
        """
        Gets the global parallelism info.
        """
        response = await self._http.request(
            path="/v2/globalParallelism",
            method="GET",
        )

        return GlobalParallelismInfo(
            parallelism_max=response.get("parallelismMax", 0),
            parallelism_count=response.get("parallelismCount", 0),
        )

    async def pause(self, flow_control_key: str) -> None:
        """
        Pause message delivery for a flow-control key.

        Messages already in the waitlist will remain there.
        New incoming messages will be added directly to the waitlist.

        :param flow_control_key: The flow control key to pause.
        """
        await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/pause",
            method="POST",
            parse_response=False,
        )

    async def resume(self, flow_control_key: str) -> None:
        """
        Resume message delivery for a flow-control key.

        :param flow_control_key: The flow control key to resume.
        """
        await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/resume",
            method="POST",
            parse_response=False,
        )

    async def pin(self, flow_control_key: str, options: PinFlowControlOptions) -> None:
        """
        Pin a processing configuration for a flow-control key.

        While pinned, the system ignores configurations provided by incoming
        messages and uses the pinned configuration instead.

        :param flow_control_key: The flow control key to pin.
        :param options: The configuration to pin.
        """
        params: Dict[str, str] = {}
        if "parallelism" in options:
            params["parallelism"] = str(options["parallelism"])
        if "rate" in options:
            params["rate"] = str(options["rate"])
        if "period" in options:
            params["period"] = str(options["period"])

        await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/pin",
            method="POST",
            params=params or None,
            parse_response=False,
        )

    async def unpin(
        self, flow_control_key: str, options: UnpinFlowControlOptions
    ) -> None:
        """
        Remove the pinned configuration for a flow-control key.

        After unpinning, the system resumes updating the configuration
        based on incoming messages.

        :param flow_control_key: The flow control key to unpin.
        :param options: Which configurations to unpin.
        """
        params: Dict[str, str] = {}
        if "parallelism" in options:
            params["parallelism"] = str(options["parallelism"]).lower()
        if "rate" in options:
            params["rate"] = str(options["rate"]).lower()

        await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/unpin",
            method="POST",
            params=params or None,
            parse_response=False,
        )

    async def reset_rate(self, flow_control_key: str) -> None:
        """
        Reset the rate configuration state for a flow-control key.

        Clears the current rate count and immediately ends the current period.
        The current timestamp becomes the start of the new rate period.

        :param flow_control_key: The flow control key to reset rate for.
        """
        await self._http.request(
            path=f"/v2/flowControl/{flow_control_key}/resetRate",
            method="POST",
            parse_response=False,
        )



from typing import List

from qstash.asyncio.http import AsyncHttpClient
from qstash.url_group import (
    RemoveEndpointRequest,
    UpsertEndpointRequest,
    UrlGroup,
    parse_url_group_response,
    prepare_add_endpoints_body,
    prepare_remove_endpoints_body,
)


class AsyncUrlGroupApi:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def upsert_endpoints(
        self,
        url_group: str,
        endpoints: List[UpsertEndpointRequest],
    ) -> None:
        """
        Add or updates an endpoint to a url group.

        If the url group or the endpoint does not exist, it will be created.
        If the endpoint exists, it will be updated.
        """
        body = prepare_add_endpoints_body(endpoints)

        await self._http.request(
            path=f"/v2/topics/{url_group}/endpoints",
            method="POST",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    async def remove_endpoints(
        self,
        url_group: str,
        endpoints: List[RemoveEndpointRequest],
    ) -> None:
        """
        Remove one or more endpoints from a url group.

        If all endpoints have been removed, the url group will be deleted.
        """
        body = prepare_remove_endpoints_body(endpoints)

        await self._http.request(
            path=f"/v2/topics/{url_group}/endpoints",
            method="DELETE",
            headers={"Content-Type": "application/json"},
            body=body,
            parse_response=False,
        )

    async def get(self, url_group: str) -> UrlGroup:
        """
        Gets the url group by its name.
        """
        response = await self._http.request(
            path=f"/v2/topics/{url_group}",
            method="GET",
        )

        return parse_url_group_response(response)

    async def list(self) -> List[UrlGroup]:
        """
        Lists all the url groups.
        """
        response = await self._http.request(
            path="/v2/topics",
            method="GET",
        )

        return [parse_url_group_response(r) for r in response]

    async def delete(self, url_group: str) -> None:
        """
        Deletes the url group and all its endpoints.
        """
        await self._http.request(
            path=f"/v2/topics/{url_group}",
            method="DELETE",
            parse_response=False,
        )



from qstash.asyncio.http import AsyncHttpClient
from qstash.signing_key import SigningKey, parse_signing_key_response


class AsyncSigningKeyApi:
    def __init__(self, http: AsyncHttpClient) -> None:
        self._http = http

    async def get(self) -> SigningKey:
        """
        Gets the current and next signing keys.
        """
        response = await self._http.request(
            path="/v2/keys",
            method="GET",
        )

        return parse_signing_key_response(response)

    async def rotate(self) -> SigningKey:
        """
        Rotates the current signing key and gets the new signing key.

        The next signing key becomes the current signing
        key, and a new signing key is assigned to the
        next signing key.
        """
        response = await self._http.request(
            path="/v2/keys/rotate",
            method="POST",
        )

        return parse_signing_key_response(response)
